/*****************************************************************/
/*                      World of GreyHawk                        */
/*      Begun in 1998, originally based on Circle3.0 and others  */
/* File:                                                         */
/* Purpose:                                                      */
/*****************************************************************/

#include "conf.h"
#include "sysdep.h"
#include <signal.h>

#include "structs.h"
#include "utils.h"
#include "comm.h"
#include "interpreter.h"
#include "handler.h"
#include "db.h"
#include "spells.h"
#include "house.h"
#include "graph.h"
#include "shop.h"
#include "map.h"
#include "events.h"


/* external variables */
extern char *room_bits[];
extern struct char_data *char_in_room(int);
extern struct char_data *character_list;
extern struct descriptor_data *descriptor_list;
extern struct room_data *world;
extern struct index_data *obj_index;
extern const char *dirs[];
extern const int rev_dir[];
extern int top_of_world;
extern int top_of_zone_table;
extern struct zone_data *zone;
extern struct zone_data *zone_table;
extern void smallmap_look(struct char_data *ch, int size, int dif_room);
extern void maplook(struct char_data *ch, sh_int size, int dif_room, int type);
extern int distance_from_shore(int room, int xcord, int ycord);
EVENTFUNC(falling_event);
EVENTFUNC(swimming_event);
void SwimmingChar(struct char_data *ch);
void DrowningChar(struct char_data *ch);
extern struct time_info_data mud_time_passed(time_t t2, time_t t1);
extern void water_log_torches(struct char_data *i);

extern void play_sound(const char *mess, struct char_data *ch, int where, int xcord, int ycord, int zcord, int type);

/*
   ch CHAR_FALLING, this routine updates status, moves them, sends appropriate
   messages, kills them if neccessary, and handles event updates.  returns
   TRUE if they are actually falling, FALSE otherwise.

   assumption 15' per room fallen:
   first check: they fall 1 room in 1 second (4 pulses)       31mph
   second check: they fall 1 room in 1/2 second (2 pulses)    43mph
   extra checks: they fall one room per pulse (1/4 second)    +8mph

   speed piles up (to 250 mph ('terminal' velocity)) since this is no longer
   a loop, they can fall 'forever' if someone sets up a 'bottomless' pit.
   This isn't completely realistic but it will suffice, they'll hit 250mph
   in 32 pulses (~8 seconds) after falling 33 rooms (about 500'). If/when
   they hit bottom, they take:  (dice((speed - number(27, 30)), 9) - dex)
   damage so falling just one room will probably not do much damage
   (4d9-dex max), but falling a long ways can be decidedly fatal.

   Note that since this is event driven, if they are summoned out, they are
   still falling that fast, and they'll take that damage.  Since this isn't
   instantaneous they can (theoretically) save themselves by starting to fly
   or levitate (we can now add feather fall that kicks in if they exceed 20
   mph (they'd still be falling though *grin*)
 */

  /* New formula; All damage taken is a percentage of their max, so a level
     1 player has just as much chance of survival as a level 50.
     max hit * (speed/2.5) + (90-120) - agil... So, one room fall, shitty char
     stats, would be 12% of their max hit, plus about 30 points, depending on
     agi. Should they reach terminal velocity, 33+ rooms, they will take 100%
     damage, + the 30 or so points. Having safe_fall skill kick in, should
     leave them living, but on the brink of death. */
void FallingChar(struct char_data *ch, const int kill_char)
{
  struct falling_event_obj *fed;

  if (!ch) {
    log("falling_char with NULL char.");
    return;
  }
  if (ch->in_room == NOWHERE)
    return;

  /* If there is a ground, assume we are here due to fall_chance. Now, if
     they can't go, its prolly due to the down exit being closed/blocked.
     Even if it is really a bug, don't bother spamming messages.
     Using virtual can go, as it checks more door states */
  if ((world[ch->in_room].sector_type != SECT_FLYING) &&
      !VIRTUAL_CAN_GO(ch->in_room, DOWN) && (GET_Z_CORD(ch) == 0))
    return;

  if ((world[ch->in_room].sector_type != SECT_FLYING) && GET_Z_CORD(ch) < 1)
    return;

  /* rooftop walking */
  if (world[IN_ROOM(ch)].map &&
      world[IN_ROOM(ch)].map->map_grid[GET_X_CORD(ch)][GET_Y_CORD(ch)] == MAP_WOOD_ROOF &&
      GET_Z_CORD(ch) == GET_Z_SIZE(ch->in_room) + 1)
    return;

  /* This isn't for sinking */
  if (GET_Z_CORD(ch) < 0)
    return;

  /* ch has just stepped into space, initiate the plunge  */
  if (IS_TRUSTED(ch) || IS_AFFECTED(ch, AFF_LEVITATE) ||
      IS_AFFECTED(ch, AFF_FLY)) {
    send_to_char("You gently float, with no solid ground touching your feet.\r\n", ch);
    return;
  }
  act("$n has just realized $e has no visible means of support!", TRUE, ch, 0, 0, PROXIMITY);
  if (GET_STAT(ch) > STAT_SLEEPING)
    send_to_char("You rediscover the law of gravity...\r\n...the hard way!\r\n", ch);
  else if (GET_STAT(ch) > STAT_INCAP)
    send_to_char("You get a sinking feeling.\r\n", ch);
  else
    send_to_char("Just when it seemed things couldn't get any worse...\r\n", ch);

  if (CAN_GO(ch, DOWN) || GET_Z_CORD(ch) > 0) {
    CREATE(fed, struct falling_event_obj, 1);
    fed->ch = ch;
    fed->obj = NULL;
    fed->speed = 0;
    fed->fatal = kill_char;

    if (GET_ACTION(ch))
      event_cancel(GET_ACTION(ch));
    GET_ACTION(ch) = NULL;
    GET_ACTION(ch) = event_create(falling_event, fed, 1);
    return;
  } else {
    send_to_char("But wait!  Saved by a bug!", ch);
    act("$n is granted a reprieve, and breathes a prayer of thanks", FALSE, ch, 0, 0, PROXIMITY);
    world[ch->in_room].sector_type = SECT_INSIDE;
    return;
  }
  return;
}


EVENTFUNC(falling_event)
{
  int dam = 0, speed, i, new_room = NOWHERE, ring = FALSE;
  bool had_zcord = FALSE, kill_char;
  struct falling_event_obj *fed = (struct falling_event_obj *) event_obj;
  struct char_data *ch;
  struct obj_data *obj;

  ch = fed->ch;
  obj = fed->obj;
  speed = fed->speed;
  kill_char = fed->fatal;


  if (ch) {
    i = ch->in_room;

    if ((GET_Z_CORD(ch) > 0) || (world[i].dir_option[DOWN] &&
	     !IS_SET(world[i].dir_option[DOWN]->exit_info, EX_CLOSED))) {
      if (GET_Z_CORD(ch) > 0) {
	GET_Z_CORD(ch)--;
	had_zcord = TRUE;
	new_room = ch->in_room;
      } else
	new_room = world[i].dir_option[DOWN]->to_room;

      if (speed < 45)
	act("$n drops from sight.", TRUE, ch, 0, 0, PROXIMITY);
      else if (speed < 90)
	act("Someone drops from sight.", TRUE, ch, 0, 0, PROXIMITY);
      else
	act("A large (screaming) object drops from sight!", TRUE, ch, 0, 0, PROXIMITY);
      play_sound("!!SOUND(fall1* Q)", ch, ch->in_room, 0, 0, GET_Z_CORD(ch), TO_ROOM);
      char_from_room(ch);
      char_to_room(ch, new_room, -2);

      /* change falling speed  */
      if (IS_AFFECTED(ch, AFF_LEVITATE) || IS_AFFECTED(ch, AFF_FLY)) {
	/* 1 gravity decel for lev, 1/2 grav for fly  */
	speed -= (IS_AFFECTED(ch, AFF_FLY)) ? 4 : 8;
	if (speed <= 0) {
	  speed = 0;
	  send_to_char("You slow to a stop.  WHEW!\r\n", ch);
	  act("$n floats in from above.", TRUE, ch, 0, 0, PROXIMITY);
	  smallmap_look(ch, view_distance(ch), -1);
	  FREE(event_obj);
	  return 0;
	} else
	  send_to_char("You slow down a little.\r\n", ch);
      } else {
	if (speed == 1)
	  speed = 31;		/* first room  */
	else if (speed == 31)
	  speed = 43;		/* second room  */
	else
	  speed += 8;		/* additional rooms  */


	if (speed > 250)
	  speed = 250;
	else if (speed < 31)
	  speed = 31;

	/* ring of feather falling */
	if ((GET_EQ(ch, WEAR_FINGER_R) && GET_OBJ_ORIG_VNUM(GET_EQ(ch, WEAR_FINGER_R)) == 2209) ||
	    (GET_EQ(ch, WEAR_FINGER_L) && GET_OBJ_ORIG_VNUM(GET_EQ(ch, WEAR_FINGER_L)) == 2209)) {
	  speed = 10;		/* slow as a feather :) */
	  ring = TRUE;
	}
      }
    } else
      new_room = i;
    /* this happens when faller is summoned/transed/teleported out */
    if (!world[new_room].dir_option[DOWN] ||
	(world[new_room].dir_option[DOWN]->to_room == NOWHERE) ||
	IS_SET(world[new_room].dir_option[DOWN]->exit_info, EX_CLOSED)
	|| ((GET_Z_CORD(ch) == 0) && had_zcord)) {

      /* oh dear, we seem to have run out of falling room!  Muhahaha  */
      dam = (int) (GET_MAX_HIT(ch) * ((speed / 2.5) / 100)) + (number(10, 20) - GET_DEX(ch));

      if (dam < 2)
	dam = 2;
#if 0
      if (GET_CHAR_SKILL(ch, SKILL_SAFE_FALL))
	if (GET_CHAR_SKILL(ch, SKILL_SAFE_FALL) > number(1, 101))
	  dam <<= 1;
#endif
      /* ring of feather falling */
      if (ring)
	dam <<= 1;

      if (dam <= 0) {
	send_to_char("You land deftly on your feet, nice jump!\r\n", ch);
	smallmap_look(ch, view_distance(ch), -1);
	act("$n drops in from above, landing neatly.", TRUE, ch, 0, 0, PROXIMITY);
	FREE(event_obj);
	return 0;
      } else if (IS_WATER_CORD(ch->in_room, GET_X_CORD(ch), GET_Y_CORD(ch))) {
	send_to_char("With a splash, you plunge into the waters!\r\n", ch);
	GET_Z_CORD(ch) = -1;
	SwimmingChar(ch);
	smallmap_look(ch, view_distance(ch), -1);
	act("$n drops in from above with a loud splash.", TRUE, ch, 0, 0, PROXIMITY);
	FREE(event_obj);
	return 0;
      } else {
	send_to_char("You land with stunning force!\r\n", ch);
	act("$n falls in from above, landing in a crumpled heap!", TRUE, ch, 0, 0, PROXIMITY);
	SET_POS(ch, number(0, 2) + GET_STAT(ch));
	if (GET_Z_CORD(ch) > 0)
	  GET_Z_CORD(ch) = 0;
	if (kill_char) {
	  damage(ch, ch, dam, TYPE_FALLING, -1, -1);
	  FREE(event_obj);
	  return 0;
	} else {
	  GET_HIT(ch) = MAX(GET_HIT(ch) - dam, -8);
	  update_pos(ch);
	}
#if 0
	Stun(ch, (100 * dam / GET_MAX_HIT(ch)));	/* 1-100  */
	/* also can knock them out for a time.  */
	if (number(1, (100 * dam / GET_MAX_HIT(ch))) >
	    number(STAT_INDEX(GET_C_CON(ch)) / 2, STAT_INDEX(GET_C_CON(ch)) * 3))
	  KnockOut(ch, number(2, MAX(2, (100 - GET_C_CON(ch)))));
#endif
	FREE(event_obj);
	return 0;
      }
    }
    /* just passing through */
    if (speed < 45) {
      act("$n falls in from above.", TRUE, ch, 0, 0, PROXIMITY);
      smallmap_look(ch, view_distance(ch), -1);
    } else if (speed < 90) {
      act("Someone hurtles in from above.", TRUE, ch, 0, 0, PROXIMITY);
    } else {
      act("A large (screaming) object plummets in from above", TRUE, ch, 0, 0, PROXIMITY);
      send_to_char("You fall, shapes and sounds shredding past you.\r\n", ch);
    }
  } else if (obj) {
    if (!obj->in_room) {
      /* somebody snagged it while it was falling */
      /* may have to do the damage here, but more likely in get() */
      FREE(event_obj);
      return 0;
    }
    if (speed < 45)
      act("$p drops from sight.", TRUE, 0, obj, 0, PROXIMITY);
    else
      act("Something drops from sight.", TRUE, 0, obj, 0, PROXIMITY);

    if (GET_OBJ_Z_CORD(obj) > 0) {
      GET_OBJ_Z_CORD(obj)--;
    } else if (world[obj->in_room].dir_option[DOWN]) {
      new_room = world[obj->in_room].dir_option[DOWN]->to_room;
      if (new_room > 1) {
	obj_from_room(obj);
	obj_to_room(obj, new_room);
	if (speed < 45)
	  act("$p falls in from above.", TRUE, 0, obj, 0, PROXIMITY);
	else
	  act("Something plummets in from above.", TRUE, 0, obj, 0, PROXIMITY);
      }
    } else if (!world[new_room].dir_option[DOWN] ||
	       (world[new_room].dir_option[DOWN]->to_room == NOWHERE) ||
	       (!GET_OBJ_Z_CORD(obj))) {
      /* we seem to have run out of falling room! */
      dam = dice((speed - number(27, 30)), 9);
      if (dam <= 0) {
	act("$p wafts to the ground.", TRUE, 0, obj, 0, PROXIMITY);
	FREE(event_obj);
	return 0;
      } else if (IS_WATER_CORD(obj->in_room, GET_OBJ_X_CORD(obj), GET_OBJ_Y_CORD(obj))) {
	GET_OBJ_Z_CORD(obj) = -(distance_from_shore(obj->in_room, GET_OBJ_X_CORD(obj), GET_OBJ_Y_CORD(obj)));
	act("$p drops in from above with a loud splash.", TRUE, 0, obj, 0, PROXIMITY);
	FREE(event_obj);
	return 0;
      } else {
	act("$p crashes into the ground.", TRUE, 0, obj, 0, PROXIMITY);
	/* add object damage/destroy, also hit people, here */
	FREE(event_obj);
	return 0;
      }
    }
    /* change falling speed */
    if (speed == 31)
      speed = 43;		/* second room */
    else
      speed += 8;		/* additional rooms */

    if (speed > 250)
      speed = 250;

  } else {
    log("SYSERR: falling_event called with no ch nor obj");
    return 0;
  }

  return (MIN(1, 100 - speed));	/* 1-100 seconds? */
}

/*
   obj OBJ_FALLING, this routine updates speed, moves it, sends appropriate
   messages, destroys/damages objects as neccessary, and handles event updates.
   returns TRUE if object is falling, FALSE otherwise.

   assumption 15' per room fallen:
   first check:  falls 1 room in 1 second (4 pulses)       31mph
   second check: falls 1 room in 1/2 second (2 pulses)    43mph
   extra checks: falls one room per pulse (1/4 second)    +8mph

   speed piles up (to 250 mph ('terminal' velocity)) since this is no longer
   a loop, it can fall 'forever' if someone sets up a 'bottomless' pit.
   This isn't completely realistic but it will suffice, it'll hit 250mph
   in 32 pulses (~8 seconds) after falling 33 rooms (about 500'). If/when
   it hits bottom, it'll take some damage (most things), or be destroyed.

   Note that since this is event driven, if object is grabbed it will xfer
   energy to the grabber, which will be in the form of increased speed (if
   grabber is also falling) and/or damage.  Same holds true if object hits
   a character (I'm not going to include falling objects hitting other objects).
 */

void FallingObj(struct obj_data *obj)
{
  struct falling_event_obj *fed;

  if (!obj) {
    log("falling_obj: NULL obj.");
    return;
  }
  /* Not for underwater use */
  if (GET_OBJ_Z_CORD(obj) <= 0)
    return;

  if (IS_OBJ_STAT(obj, ITEM_LEVITATES))
    return;

  if (obj->carried_by || obj->worn_by)
    return;

  /* obj has just been dropped/moved into space, initiate the plunge */
  act("$p drops from sight.", TRUE, 0, obj, 0, PROXIMITY);

  CREATE(fed, struct falling_event_obj, 1);
  fed->ch = NULL;
  fed->obj = obj;
  fed->speed = 31;
  fed->fatal = 0;
  GET_OBJ_EVENT(obj) = NULL;
  GET_OBJ_EVENT(obj) = event_create(falling_event, fed, 1);
  return;
}

struct swimming_event_obj {
  struct char_data *ch;
  time_t began_swim;
};

/* Called upon first entry to water. This sets bits, then calls
   the event which handles the actual maintenance of swimmers */
void SwimmingChar(struct char_data *ch)
{
  struct swimming_event_obj *sed;

  if (!ch)
    return;

  if (IS_TRUSTED(ch) || GET_Z_CORD(ch) > 0 || IS_AFFECTED(ch, AFF_LEVITATE)
      || IS_AFFECTED(ch, AFF_FLY))
    if (IS_SET(AFF3_FLAGS(ch), AFF3_SWIMMING) && !IS_SWIMMING_RACE(ch))
      REMOVE_BIT(AFF3_FLAGS(ch), AFF3_SWIMMING);

  if (GET_STAT(ch) == STAT_DEAD) {
    log("SYSERR: dead char in call to SwimmingChar()");
    raise(SIGSEGV);
  }
  if (IN_ROOM(ch) == NOWHERE || !IS_WATER_CORD(IN_ROOM(ch), GET_X_CORD(ch), GET_Y_CORD(ch))) {
    if (IS_SWIMMING_RACE(ch))
      vlogf("SYSERR: %s was messed up in SwimmingChar.", GET_NAME(ch));
    if (IS_SET(AFF2_FLAGS(ch), AFF2_HOLDING_BREATH))
      REMOVE_BIT(AFF2_FLAGS(ch), AFF2_HOLDING_BREATH);
    if (IS_SET(AFF2_FLAGS(ch), AFF2_IS_DROWNING))
      REMOVE_BIT(AFF2_FLAGS(ch), AFF2_IS_DROWNING);
    if (IS_SET(AFF3_FLAGS(ch), AFF3_SWIMMING))
      REMOVE_BIT(AFF3_FLAGS(ch), AFF3_SWIMMING);
    return;
  }
  if (IS_SWIMMING_RACE(ch)) {
    if (IS_SET(AFF3_FLAGS(ch), AFF3_SWIMMING) && GET_Z_CORD(ch) >= 0)
      GET_Z_CORD(ch) = -1;	/* keep fish in water */
    return;
  }
  if (IS_SET(AFF2_FLAGS(ch), AFF2_HOLDING_BREATH) ||
      IS_SET(AFF2_FLAGS(ch), AFF2_IS_DROWNING) ||
      (IS_SET(AFF3_FLAGS(ch), AFF3_SWIMMING) && GET_ACTION(ch)))
    return;

  SET_BIT(AFF3_FLAGS(ch), AFF3_SWIMMING);
  play_sound("!!SOUND(splash* Q)", ch, ch->in_room, 0, 0, GET_Z_CORD(ch), TO_ROOM);
  send_to_char("You begin to swim...\r\n", ch);
  act("$n begins to swim.", TRUE, ch, 0, 0, TO_PROXIMITY);

  water_log_torches(ch);

  if (GET_ACTION(ch))
    event_cancel(GET_ACTION(ch));
  CREATE(sed, struct swimming_event_obj, 1);
  sed->ch = ch;
  sed->began_swim = time(0);
  GET_ACTION(ch) = NULL;
  GET_ACTION(ch) = event_create(swimming_event, sed, GET_CON(ch));
  return;
}

/* mostly just keeps track of the time in the water, setting the player
   up for drowning once they've tired out. Also clears itself if the
   danger is removed. */
EVENTFUNC(swimming_event)
{
  struct swimming_event_obj *sed = (struct swimming_event_obj *) event_obj;
  struct char_data *ch;
  struct time_info_data swim_time;

  ch = sed->ch;
  swim_time = mud_time_passed(time(0), sed->began_swim);

  if (IN_ROOM(ch) == NOWHERE || !IS_WATER_CORD(IN_ROOM(ch), GET_X_CORD(ch), GET_Y_CORD(ch))) {
    if (IS_SET(AFF2_FLAGS(ch), AFF2_HOLDING_BREATH))
      REMOVE_BIT(AFF2_FLAGS(ch), AFF2_HOLDING_BREATH);
    if (IS_SET(AFF2_FLAGS(ch), AFF2_IS_DROWNING))
      REMOVE_BIT(AFF2_FLAGS(ch), AFF2_IS_DROWNING);
    if (IS_SET(AFF3_FLAGS(ch), AFF3_SWIMMING))
      REMOVE_BIT(AFF3_FLAGS(ch), AFF3_SWIMMING);
    FREE(event_obj);
    return 0;
  }
  /* Normal swimmer */
  if (IS_SET(AFF3_FLAGS(ch), AFF3_SWIMMING)) {
    /* getting tired */
    if (swim_time.minutes > GET_CON(ch) * 2) {
      FREE(event_obj);		/* free this one now, because we are adding another */
      REMOVE_BIT(AFF3_FLAGS(ch), AFF3_SWIMMING);
      if (GET_Z_CORD(ch) == 0)
	GET_Z_CORD(ch) = -1;
      DrowningChar(ch);
      return 0;
    }
  }
  /* Out of air swimmer */
  if (IS_SET(AFF2_FLAGS(ch), AFF2_HOLDING_BREATH)) {
    if (swim_time.minutes > GET_CON(ch) / 2) {
      REMOVE_BIT(AFF3_FLAGS(ch), AFF3_SWIMMING);
      REMOVE_BIT(AFF2_FLAGS(ch), AFF2_HOLDING_BREATH);
      SET_BIT(AFF2_FLAGS(ch), AFF2_IS_DROWNING);
      if (GET_Z_CORD(ch) == 0)
	GET_Z_CORD(ch) = -1;
    } else {
      send_to_char("Bloop... bloop... bloop...\r\n", ch);
    }
  }
  /* Drowning char */
  if (IS_SET(AFF2_FLAGS(ch), AFF2_IS_DROWNING)) {
    send_to_char("The lack of air is becoming a serious problem!\r\n", ch);
    act("$n would be gasping for air, if $e weren't DROWNING!", TRUE, ch, 0, 0, TO_PROXIMITY);
    if (GET_HIT(ch) > 5)
      GET_HIT(ch) -= 3;
    else
      damage(ch, ch, 3, TYPE_DROWNING, -1, -1);
  }
  return GET_CON(ch);

}

/* Checks if drowning conditions are still valid, calls event if so */
void DrowningChar(struct char_data *ch)
{
  struct swimming_event_obj *sed;

  if (!ch)
    return;

  if (GET_RACE(ch) == RACE_ELEMENTAL_WATER || GET_RACE(ch) == RACE_FISH)
    return;

  /* ring of swimmming */
  if ((GET_EQ(ch, WEAR_FINGER_R) && GET_OBJ_ORIG_VNUM(GET_EQ(ch, WEAR_FINGER_R)) == 2229) ||
      (GET_EQ(ch, WEAR_FINGER_L) && GET_OBJ_ORIG_VNUM(GET_EQ(ch, WEAR_FINGER_L)) == 2229))
    return;

  if (IS_TRUSTED(ch) || GET_Z_CORD(ch) > 0 || IS_AFFECTED(ch, AFF_LEVITATE)
      || IS_AFFECTED(ch, AFF_FLY) || IS_AFFECTED(ch, AFF_WATERBREATH)) {
    if (IS_SET(AFF2_FLAGS(ch), AFF2_HOLDING_BREATH))
      REMOVE_BIT(AFF2_FLAGS(ch), AFF2_HOLDING_BREATH);
    if (IS_SET(AFF2_FLAGS(ch), AFF2_IS_DROWNING))
      REMOVE_BIT(AFF2_FLAGS(ch), AFF2_IS_DROWNING);
    return;
  }
  if (GET_STAT(ch) == STAT_DEAD) {
    log("SYSERR: dead char in call to DrowningChar()");
    return;
  }
  if (IN_ROOM(ch) == NOWHERE || !IS_WATER_CORD(IN_ROOM(ch), GET_X_CORD(ch), GET_Y_CORD(ch))) {
    if (IS_SET(AFF2_FLAGS(ch), AFF2_HOLDING_BREATH))
      REMOVE_BIT(AFF2_FLAGS(ch), AFF2_HOLDING_BREATH);
    if (IS_SET(AFF2_FLAGS(ch), AFF2_IS_DROWNING))
      REMOVE_BIT(AFF2_FLAGS(ch), AFF2_IS_DROWNING);
    if (IS_SET(AFF3_FLAGS(ch), AFF3_SWIMMING))
      REMOVE_BIT(AFF3_FLAGS(ch), AFF3_SWIMMING);
    return;
  }
  if (!IS_SET(AFF2_FLAGS(ch), AFF2_HOLDING_BREATH) &&
      !IS_SET(AFF2_FLAGS(ch), AFF2_IS_DROWNING)) {
    SET_BIT(AFF2_FLAGS(ch), AFF2_HOLDING_BREATH);
    send_to_char("You begin to hold your breath.../r/n", ch);
    act("$n looks like $e's having troubles with the air supply.", TRUE, ch, 0, 0, TO_PROXIMITY);
  }
  if (GET_ACTION(ch))
    event_cancel(GET_ACTION(ch));
  CREATE(sed, struct swimming_event_obj, 1);
  sed->ch = ch;
  sed->began_swim = time(0);
  GET_ACTION(ch) = NULL;
  GET_ACTION(ch) = event_create(swimming_event, sed, GET_CON(ch));
  return;
}
