
/*****************************************************************/
/*                      World of GreyHawk                        */
/*      Begun in 1998, originally based on Circle3.0 and others  */
/* File:                                                         */
/* Purpose:                                                      */
/*****************************************************************/
/**************************************************************************
* Using the memory debugging                                              *
*                                                                         *
* The memory debugging option keeps a list of all memory allocated with   *
* the CREATE or RECREATE macros, and which file and line it was allocated *
* at.  FREE removes the entry from the list.  When the game is shut down, *
* a list of all memory not FREEd, which file and line it was created, and *
* how many bytes will be recorded to log/memory.log.  It will also log    *
* any attempts to FREE an object not in its list of allocated memory,     *
* for example, memory already FREEd or memory not alloced with CREATE or  *
* RECREATE.                                                               *
* Optionally, all allocations and frees can be logged.                    *
*                                                                         *
* To turn on this feature, define MEMCHK and recompile.  Defining         *
* MEMCHK = 2 will turn on the additional logging of the allocs and frees  *
**************************************************************************/
#include "conf.h"
#include "sysdep.h"
#include <signal.h>

#include "structs.h"
#include "utils.h"

struct mem_list {
  void *loc;
  size_t size;
  char *file;
  int line;
  struct mem_list *next;
};

#define MEMORY_LOG "../log/memory"

#ifdef MEMCHK

FILE *mem_log = NULL;
struct mem_list *memory_list = NULL;	/* The list of memory alloced so far */


/* adds a piece of memory to the list */
void addmem(void *p, size_t size, char *file, int line)
{
  struct mem_list *l;
  time_t ct;
  char *tmstr, str[80];

  /* open the memory log if not already */
  if (!mem_log) {
    if ((mem_log = fopen(MEMORY_LOG, "w")) == NULL) {
      /* IH 05-oct-98 FIX:make error more meaningful */
      log("addmem:fopen memory log");
      raise(SIGSEGV);
    }
  }
#if MEMCHK > 1
  sprintf(str, "%p: Allocating %d bytes, file %s:%d",
	  p, size, file, line);
  ct = time(0);
  tmstr = asctime(localtime(&ct));
  *(tmstr + strlen(tmstr) - 1) = '\0';
  fprintf(mem_log, "%-19.19s :: %s\n", tmstr, str);
#endif

  if ((l = (struct mem_list *) malloc(sizeof(struct mem_list))) == NULL) {
    log("malloc");
    raise(SIGSEGV);
  }
  /* init new node */
  l->loc = p;
  l->size = size;
  l->file = file;
  l->line = line;

  /* add node to list */
  l->next = memory_list;
  memory_list = l;
}


/* removes a memory entry from the list */
void delmem(void *p, char *file, int line)
{
  struct mem_list *l, *m;
  time_t ct;
  char *tmstr, str[80];

  /* open the memory log if not already */
  if (!mem_log) {
    if ((mem_log = fopen(MEMORY_LOG, "w")) == NULL) {
      /* IH 05-oct-98 FIX:make error more meaningful */
      log("delmem:fopen memory log");
      raise(SIGSEGV);
    }
  }
  /* find the entry in the list */
  for (l = NULL, m = memory_list;
       m && (m->loc != p);
       l = m, m = m->next);

  if (m) {
#if MEMCHK > 1
    sprintf(str, "Freed in file %s: %d: %p", file, line, p);
    ct = time(0);
    tmstr = asctime(localtime(&ct));
    *(tmstr + strlen(tmstr) - 1) = '\0';
    fprintf(mem_log, "%-19.19s :: %s\n", tmstr, str);
#endif

    if (m == memory_list) {	/* head of list */
      memory_list = m->next;
      free((char *) m);
    } else {
      l->next = m->next;
      free((char *) m);
    }
  } else
    fprintf(mem_log, "Attempting to free non-alloced memory, file %s:%d\n",
	    file, line);
}


/* writes the list of all unfreed memory to file */
void dump_mem_log(void)
{
  struct mem_list *l, *next;
  size_t total = 0;

  /* open the memory log if not already */
  if (!mem_log) {
    if ((mem_log = fopen(MEMORY_LOG, "w")) == NULL) {
      /* IH 05-oct-98 FIX:make error more meaningful */
      log("dump_mem_log:fopen memory log");
      raise(SIGSEGV);
    }
  }
  for (l = memory_list; l; l = next) {
    next = l->next;
    total += l->size;
    fprintf(mem_log, "%d bytes of unfreed memory at %p, %s:%d\n", l->size, l,
	    l->file, l->line);
    free((char *) l);
  }

  fprintf(mem_log, "Total unfreed memory: %d\n", total);

  fclose(mem_log);
}


void *debug_calloc(size_t nobj, size_t size, char *file, int line)
{
  void *m;

  m = calloc(nobj, size);

  if (m)
    addmem(m, nobj * size, file, line);

  return m;
}


void *debug_realloc(void *p, size_t size, char *file, int line)
{
  void *m;

  delmem(p, file, line);

  m = realloc(p, size);

  if (m)
    addmem(m, size, file, line);

  return m;
}


void debug_free(void *p, char *file, int line)
{
  delmem(p, file, line);

  free(p);
}

#endif				/* MEMCHK */
